﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.IO;
using System.Threading;

namespace SalesVizu
{

    public class Threading
    {

        /// <summary>
        /// A threaded task
        /// </summary>
        public delegate void ThreadTaskDelegate();

        /// <summary>
        /// A method to be called when a task finishes
        /// </summary>
        /// <param name="taskId">The task identifier for this task (-1 by default)</param>
        public delegate void TaskFinishedDelegate(int taskId);

        /// <summary>
        /// A task to thread
        /// </summary>
        public struct ThreadTask
        {

            /// <summary>
            /// The task this thread runs
            /// </summary>
            private ThreadTaskDelegate task;

            /// <summary>
            /// Gets the task this thread runs
            /// </summary>
            public ThreadTaskDelegate Task { get { return task; } }

            /// <summary>
            /// The method called when the task is finished
            /// </summary>
            private TaskFinishedDelegate taskFinished;

            /// <summary>
            /// Gets the task finished method
            /// </summary>
            public TaskFinishedDelegate TaskFinished { get { return taskFinished; } }

            /// <summary>
            /// A task identifier (-1 by default)
            /// </summary>
            private int taskId;

            /// <summary>
            /// Gets the task identifier of this task (-1 by default)
            /// </summary>
            public int TaskId { get { return taskId; } }


            /// <summary>
            /// Initializes a thread task
            /// </summary>
            /// <param name="task">the method to call</param>
            /// <param name="taskFinished">The method to call when finished</param>
            /// <param name="taskId">An identifier for this task</param>
            public ThreadTask(
                ThreadTaskDelegate task,
                TaskFinishedDelegate taskFinished,
                int taskId)
            {
                this.task = task;
                this.taskFinished = taskFinished;
                this.taskId = taskId;
            }

            /// <summary>
            /// Initializes a thread task
            /// </summary>
            /// <param name="task">the method to call</param>
            /// <param name="taskFinished">The method to call when finished</param>
            public ThreadTask(
                ThreadTaskDelegate task,
                TaskFinishedDelegate taskFinished) :
                this(task, taskFinished, -1)
            {
            }

            /// <summary>
            /// Initializes a thread task
            /// </summary>
            /// <param name="task">the method to call</param>
            public ThreadTask(ThreadTaskDelegate task)
                : this(task, null, -1)
            {
            }
        }


        public class ManagedThreadEveryFrame : ManagedThread
        {
            AutoResetEvent frameStart, completion;
            public GameTime gameTime;

            public ManagedThreadEveryFrame(int processorAffinity)
                : base(
#if XBOX
                processorAffinity
#endif
)
            {
                frameStart = new AutoResetEvent(false);
                completion = new AutoResetEvent(false);
            }

            public void FrameStart(GameTime gameTime)
            {
                this.gameTime = gameTime;
                frameStart.Set();
            }
            public void WaitForFrameStart() { frameStart.WaitOne(); }
            public void WaitForCompletion() { completion.WaitOne(); }
            public void Completed() { completion.Set(); }
        }

        /// <summary>
        /// A managed thread
        /// </summary>
        public class ManagedThread
        {
            static List<ManagedThread> instances = new List<ManagedThread>();
            public static void KillAll()
            {
                foreach (ManagedThread t in instances)
                {
                    if (t != null)
                    {
                        t.KillImmediately();
                    }
                }
            }

            /// <summary>
            /// The thread managed by this object
            /// </summary>
            private Thread thread;

            /// <summary>
            /// The processor this thread should run on
            /// </summary>
            private int processorAffinity;

            /// <summary>
            /// True when this thread should be killed
            /// </summary>
            private bool killThread = false;

            /// <summary>
            /// Tasks to be run by this thread
            /// </summary>
            private Queue<ThreadTask> tasks = new Queue<ThreadTask>();


            /// <summary>
            /// Initializes the thread with a specified processor affinity
            /// </summary>
            public ManagedThread(int processorAffinity,
#if WINDOWS
 ThreadPriority prio,
#endif
 bool registerInThreadList
                )
            {
                this.processorAffinity = processorAffinity;
#if XBOX
                if(processorAffinity != -1)
                {
                    Thread.CurrentThread.SetProcessorAffinity(new int[] { processorAffinity });
                }
#endif

                // Create the starter object
                ThreadStart starter = new ThreadStart(taskRunner);
                thread = new Thread(starter);
#if WINDOWS
                thread.Priority = prio;
#endif
                thread.Start();

                if (registerInThreadList)
                {
                    instances.Add(this);
                }
            }

#if WINDOWS
            public ManagedThread(ThreadPriority prio)
                : this(-1, prio, true)
            { }
            public ManagedThread(int processorAffinity)
                : this()
            { }
            public ManagedThread()
                : this(ThreadPriority.Normal)
            { }
#else
            public ManagedThread()
                : this(-1)
            { }
            public ManagedThread(int processorAffinity)
                : this(processorAffinity, true)
            { }

#endif



            /*
            public void WaitForSignal()
            {
                waiting = true;
            }
             */

            /// <summary>
            /// Runs tasks when there are any to run
            /// </summary>
            private void taskRunner()
            {
                #region Set processor
#if XBOX
                // Ensure processor affinity is within range (1, 3, 4 or 5)
                if (processorAffinity > 0 && processorAffinity < 6 &&
                    processorAffinity != 2)
                {
                    // Set the processor affinity
                    thread.SetProcessorAffinity(new int[] { processorAffinity });
                }	
#endif
                #endregion

                #region Run Tasks

                // The task to run
                ThreadTask task;

                while (!killThread)
                {
                    if (tasks.Count > 0)
                    {
                        // Run the task
                        lock (tasks)
                        {
                            task = tasks.Dequeue();
                        }
                        task.Task();
                        if (task.TaskFinished != null)
                        {
                            task.TaskFinished(task.TaskId);
                        }
                    }
                    else
                    {
                        // Wait for more tasks to run
                        Thread.Sleep(0);
                    }
                }

                #endregion

                // Clean up
                tasks.Clear();
                tasks = null;
            }

            /// <summary>
            /// Kills this thread, stopping at next safe point
            /// </summary>
            public void Kill()
            {
                killThread = true;
            }

            /// <summary>
            /// Kills the thread immediately
            /// </summary>
            public void KillImmediately()
            {
                killThread = true;
                thread.Abort();
            }

            /// <summary>
            /// Adds a task to this thread
            /// </summary>
            /// <param name="task">The task to perform</param>
            public void AddTask(ThreadTask task)
            {
                lock (tasks)
                {
                    tasks.Enqueue(task);
                }
            }

        }

        // obsolete : 1er essai
#if false
            public BackgroundLoading(GeoGame.UserData userData)
            {
                this.userData = userData;
            }

            // access to data            
            GeoGame.UserData userData;

            // User Settings
            ThreadState stateUserSettings = new ThreadState();
            void callBackUserSettings()
            {
                stateUserSettings.Finished = true;
            }
            public void LoadUserSettings()
            {
                (new Thread(
            }
#endif

    }


    public class LibMaximinus
    {
        public static void CrashCatcher(Game game)
        {
#if WINDOWS // && !DEBUG
            if (System.Diagnostics.Debugger.IsAttached == false)
            {
                try
                {
                    game.Run();
                }
                catch (Exception e)
                {
                    String tmpdir = System.Environment.GetEnvironmentVariable("TMP");
                    String filename = tmpdir + "\\crashlog.txt";
                    StreamWriter sw = new StreamWriter(filename);
                    sw.WriteLine("crash date : " + DateTime.Now + System.Environment.NewLine);
                    sw.WriteLine("crash log file : " + filename + System.Environment.NewLine);
                    sw.WriteLine("crash message : " + e.ToString() + " / " + e.Message + System.Environment.NewLine);
                    sw.Close();
                    System.Diagnostics.Process p = new System.Diagnostics.Process();
                    p.StartInfo.FileName = filename;
                    p.Start();
                    //System.Diagnostics.Debugger.Break();
                    BreakDebugger();
                }
            }
            else
            {
                game.Run();
            }
#else
            game.Run();
#endif
        }
        public static void BreakDebugger()
        {
            if (System.Diagnostics.Debugger.IsAttached)
            {
                try
                {
                    System.Diagnostics.Debugger.Break();
                }
                catch (Exception e)
                {
                    Utils.DebugOut(e);
                }
            }
        }
        public static RenderTarget2D NewRenderTarget(int width, int height)
        {
            return new RenderTarget2D(
                MaximinusGame.Instance.GraphicsDevice,
                width, height,
                false,
                MaximinusGame.Instance.GraphicsDevice.PresentationParameters.BackBufferFormat,
                MaximinusGame.Instance.GraphicsDevice.PresentationParameters.DepthStencilFormat,
                MaximinusGame.Instance.GraphicsDevice.PresentationParameters.MultiSampleCount,
                MaximinusGame.Instance.GraphicsDevice.PresentationParameters.RenderTargetUsage);
        }

        public class MultiMonitorGraphicsDeviceManager : GraphicsDeviceManager
        {
            int MonitorIndex = 1;

            public MultiMonitorGraphicsDeviceManager(Game game, int monitorIndex)
                : base(game)
            {
                this.MonitorIndex = monitorIndex;
            }

            protected override void OnPreparingDeviceSettings(object sender, PreparingDeviceSettingsEventArgs args)
            {
                //args.GraphicsDeviceInformation.PresentationParameters.FullScreenRefreshRateInHz = 0;
                //args.GraphicsDeviceInformation.PresentationParameters.IsFullScreen = fullscreen;
                //args.GraphicsDeviceInformation.PresentationParameters.BackBufferWidth = W;
                //args.GraphicsDeviceInformation.PresentationParameters.BackBufferHeight = H;

                base.OnPreparingDeviceSettings(sender, args);
            }


#if WINDOWS
            protected override void RankDevices(List<GraphicsDeviceInformation> foundDevices)
            {
                List<GraphicsDeviceInformation> removals = new List<GraphicsDeviceInformation>();
                for (int i = 0; i < foundDevices.Count; i++)
                {
                    if (!foundDevices[i].Adapter.DeviceName.Contains("DISPLAY" + MonitorIndex))
                        removals.Add(foundDevices[i]);
                }
                foreach (GraphicsDeviceInformation info in removals)
                {
                    foundDevices.Remove(info);
                }

                base.RankDevices(foundDevices);
            }
#endif
        }

    }
    public class MaximinusGame : Game
    {
        public static MaximinusGame Instance;
        public SpriteBatch SpriteBatch;
        public SpriteFont Font;
        public Texture2D BlankTex;
        //public static ContentM
        public static int CurrentFrame = 0;
        public Point ScreenSizePoint;
        public static bool ContentLoadFinished = false;

        public MaximinusGame(
            int width,
            float aspect)
        {
            GraphicsDeviceManager dm = new LibMaximinus.MultiMonitorGraphicsDeviceManager(this, 1);
            dm.PreferredBackBufferWidth = width;
            dm.PreferredBackBufferHeight = (int)(width / aspect) + 1;
            dm.PreferMultiSampling = false;
            dm.IsFullScreen = false;

            Instance = this;
        }

        protected override void Initialize()
        {

            ScreenSizePoint = new Point(
                GraphicsDevice.Viewport.Width,
                GraphicsDevice.Viewport.Height);

            Font = Content.Load<SpriteFont>("Fonts/font720");
            SpriteBatch = new SpriteBatch(GraphicsDevice);
            BlankTex = new Texture2D(GraphicsDevice, 1, 1);
            BlankTex.SetData(new Color[] { Color.White });

            base.Initialize();
        }

        protected override void Update(GameTime gameTime)
        {
            CurrentFrame++;
            base.Update(gameTime);
        }


        public static void LaunchSystemCommand(String fileName, String arguments, bool waitForEnd, bool hideOutput)
        {
            System.Diagnostics.Process p = new System.Diagnostics.Process();
            p.StartInfo.FileName = fileName;
            if (arguments != null)
            {
                p.StartInfo.Arguments = arguments;
            }

            if (hideOutput)
            {
                p.StartInfo.CreateNoWindow = true;
                p.StartInfo.UseShellExecute = false;
                p.StartInfo.RedirectStandardOutput = true;
            }
            p.Start();
            if (waitForEnd)
            {
                p.WaitForExit();
            }
        }

        public static void DrawStringOutlined(
            String text,
            SpriteFont font,
            Vector2 pos,
            Color faceColor,
            float scale,
            int outlineWidth)
        {
            pos = new Vector2((int)pos.X, (int)pos.Y);
            Vector2 origin = font.MeasureString(text) / 2;
            origin = new Vector2((int)origin.X, (int)origin.Y);
            for (int x = -outlineWidth; x <= +outlineWidth; x++)
            {
                for (int y = -outlineWidth; y <= +outlineWidth; y++)
                {
                    if (x != y || x != 0)
                    {
                        Instance.SpriteBatch.DrawString(font, text, pos + new Vector2(x, y), Color.Black, 0, origin, scale, SpriteEffects.None, 0);
                    }
                }
            }
            Instance.SpriteBatch.DrawString(font, text, pos, faceColor, 0, origin, scale, SpriteEffects.None, 0);
        }
        public static void DrawStringOutlinedNotCentered(String text, Vector2 pos, Color faceColor, float scale, int outlineWidth)
        {
            pos = new Vector2((int)pos.X, (int)pos.Y);
            for (int x = -outlineWidth; x <= +outlineWidth; x++)
            {
                for (int y = -outlineWidth; y <= +outlineWidth; y++)
                {
                    if (x != y || x != 0)
                    {
                        Instance.SpriteBatch.DrawString(Instance.Font, text, pos + new Vector2(x, y), Color.Black, 0, Vector2.Zero, scale, SpriteEffects.None, 0);
                    }
                }
            }
            Instance.SpriteBatch.DrawString(Instance.Font, text, pos, faceColor, 0, Vector2.Zero, scale, SpriteEffects.None, 0);
        }
        public static void OutlineRectangle(Rectangle rect, Color color, int width)
        {
            Instance.SpriteBatch.Draw(Instance.BlankTex, new Rectangle(rect.Left, rect.Top, rect.Width, width), color);
            Instance.SpriteBatch.Draw(Instance.BlankTex, new Rectangle(rect.Left, rect.Bottom - width, rect.Width, width), color);

            Instance.SpriteBatch.Draw(Instance.BlankTex, new Rectangle(rect.Left, rect.Top + width, width, rect.Height - 2 * width), color);
            Instance.SpriteBatch.Draw(Instance.BlankTex, new Rectangle(rect.Right - width, rect.Top + width, width, rect.Height - 2 * width), color);
        }


    }
    public class Utils
    {
        public static readonly String newLine = "\n";
        public static Point Vec2ToPoint(Vector2 v) { return new Point((int)v.X, (int)v.Y); }
        public static Vector2 PointToVec2(Point v) { return new Vector2(v.X, v.Y); }
        public static void SetAndClearRender(GraphicsDevice dev, RenderTarget2D r, Color c)
        {
            dev.SetRenderTarget(r);
            dev.Clear(c);
        }
        public static void DebugOut(String s)
        {
            Console.WriteLine("[" + MaximinusGame.CurrentFrame + "] " + s);
        }
        public static void DebugOut(Exception e)
        {
            DebugOut(e.ToString());
        }
    }
}
